# Copyright (C) 2002-2022
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import traceback
import re
import os
from PyQt5.Qt import  QValidator, QWidget, QHBoxLayout, QFileDialog, QComboBox, QLabel, \
    QLineEdit, QDialog, QIntValidator, QDoubleValidator, QToolButton, QListWidget, QVBoxLayout, \
    QPushButton, QMessageBox
from PyQt5.QtGui import QMouseEvent, QFont
from PyQt5.QtWidgets import QDockWidget, QStyle, QSizePolicy, QToolTip, QApplication
from PyQt5.Qwt import QwtPlotCurve, QwtPlotMarker
from PyQt5.QtCore import pyqtSignal, Qt, QEvent, QSize, QTimer

from TigGUI.init import pixmaps

QStringList = list


class TiggerPlotCurve(QwtPlotCurve):
    """Wrapper around QwtPlotCurve to make it compatible with numpy float types"""

    def setData(self, x, y):
        return QwtPlotCurve.setSamples(self, list(map(float, x)), list(map(float, y)))

    def setDataXfy(self, x, y):
        return QwtPlotCurve.setSamples(self, list(map(float, y)), list(map(float, x)))


class TiggerPlotMarker(QwtPlotMarker):
    """Wrapper around QwtPlotCurve to make it compatible with numpy float types"""

    def setValue(self, x, y):
        return QwtPlotMarker.setValue(self, float(x), float(y))


class FloatValidator(QValidator):
    """QLineEdit validator for float items in standard or scientific notation"""
    re_intermediate = re.compile("^-?([0-9]*)\.?([0-9]*)([eE]([+-])?[0-9]*)?$")

    def validate(self, _input, _pos):
        _input = str(_input)
        try:
            x = float(_input)
            return QValidator.Acceptable, _input, _pos
        except:
            pass
        if not _input or self.re_intermediate.match(_input):
            return QValidator.Intermediate, _input, _pos
        # return QValidator.Invalid, pos  # old line
        return QValidator.Acceptable, _input, _pos


class ValueTypeEditor(QWidget):
    ValueTypes = (bool, int, float, complex, str)

    def __init__(self, *args):
        QWidget.__init__(self, *args)
        lo = QHBoxLayout(self)
        lo.setContentsMargins(0, 0, 0, 0)
        lo.setSpacing(5)
        # type selector
        self.wtypesel = QComboBox(self)
        for i, tp in enumerate(self.ValueTypes):
            self.wtypesel.addItem(tp.__name__)
        self.wtypesel.activated[int].connect(self._selectTypeNum)
        typesel_lab = QLabel("&Type:", self)
        typesel_lab.setBuddy(self.wtypesel)
        lo.addWidget(typesel_lab, 0)
        lo.addWidget(self.wtypesel, 0)
        self.wvalue = QLineEdit(self)
        self.wvalue_lab = QLabel("&Value:", self)
        self.wvalue_lab.setBuddy(self.wvalue)
        self.wbool = QComboBox(self)
        self.wbool.addItems(["false", "true"])
        self.wbool.setCurrentIndex(1)
        lo.addWidget(self.wvalue_lab, 0)
        lo.addWidget(self.wvalue, 1)
        lo.addWidget(self.wbool, 1)
        self.wvalue.hide()
        # make input validators
        self._validators = {int: QIntValidator(self), float: QDoubleValidator(self)}
        # select bool type initially
        self._selectTypeNum(0)

    def _selectTypeNum(self, index):
        tp = self.ValueTypes[index]
        self.wbool.setVisible(tp is bool)
        self.wvalue.setVisible(tp is not bool)
        self.wvalue_lab.setBuddy(self.wbool if tp is bool else self.wvalue)
        self.wvalue.setValidator(self._validators.get(tp, None))

    def setValue(self, value):
        """Sets current value"""
        for i, tp in enumerate(self.ValueTypes):
            if isinstance(value, tp):
                self.wtypesel.setCurrentIndex(i)
                self._selectTypeNum(i)
                if tp is bool:
                    self.wbool.setCurrentIndex(1 if value else 0)
                else:
                    self.wvalue.setText(str(value))
                return
        # unknown value: set bool
        self.setValue(True)

    def getValue(self):
        """Returns current value, or None if no legal value is set"""
        tp = self.ValueTypes[self.wtypesel.currentIndex()]
        if tp is bool:
            return bool(self.wbool.currentIndex())
        else:
            try:
                return tp(self.wvalue.text())
            except:
                print("Error converting input to type ", tp.__name__)
                traceback.print_exc()
                return None


class FileSelector(QWidget):
    """A FileSelector is a one-line widget for selecting a file."""
    valid = pyqtSignal()
    filenameSelected = pyqtSignal()

    def __init__(self, parent, label, filename=None, dialog_label=None, file_types=None, default_suffix=None,
                 file_mode=QFileDialog.AnyFile):
        QWidget.__init__(self, parent)
        lo = QHBoxLayout(self)
        lo.setContentsMargins(0, 0, 0, 0)
        lo.setSpacing(5)
        # label
        lab = QLabel(label, self)
        lo.addWidget(lab, 0)
        # text field
        self.wfname = QLineEdit(self)
        self.wfname.setReadOnly(True)
        self.setFilename(filename)
        lo.addWidget(self.wfname, 1)
        # selector
        wsel = QToolButton(self)
        wsel.setText("Choose...")
        wsel.clicked.connect(self._chooseFile)
        lo.addWidget(wsel, 0)
        # other init
        self._file_dialog = None
        self._dialog_label = dialog_label or label
        self._file_types = file_types or "All files (*)"
        self._file_mode = file_mode
        self._default_suffix = default_suffix
        self._dir = None

    def _chooseFile(self):
        if self._file_dialog is None:
            dialog = self._file_dialog = QFileDialog(self, self._dialog_label, ".", self._file_types)
            if self._default_suffix:
                dialog.setDefaultSuffix(self._default_suffix)
            dialog.setFileMode(self._file_mode)
            dialog.setModal(True)
            if self._dir is not None:
                dialog.setDirectory(self._dir)
            dialog.filesSelected['QStringList'].connect(self.setFilename)
        return self._file_dialog.exec_()

    def setFilename(self, filename):
        if isinstance(filename, QStringList):
            filename = filename[0]
        filename = (filename and str(filename)) or ''
        self.wfname.setText(filename)
        self.valid.emit(bool(filename))
        self.filenameSelected.emit(filename)

    def setDirectory(self, directory):
        self._dir = directory
        if self._file_dialog is not None:
            self._file_dialog.setDirectory(directory)

    def filename(self):
        return str(self.wfname.text())

    def isValid(self):
        return bool(self.filename())


class AddTagDialog(QDialog):
    def __init__(self, parent, modal=True, flags=Qt.WindowFlags()):
        QDialog.__init__(self, parent, flags)
        self.setModal(modal)
        self.setWindowTitle("Add Tag")
        lo = QVBoxLayout(self)
        lo.setContentsMargins(10, 10, 10, 10)
        lo.setSpacing(5)
        # tag selector
        lo1 = QHBoxLayout()
        lo.addLayout(lo1)
        lo1.setSpacing(5)
        self.wtagsel = QComboBox(self)
        self.wtagsel.setEditable(True)
        wtagsel_lbl = QLabel("&Tag:", self)
        wtagsel_lbl.setBuddy(self.wtagsel)
        lo1.addWidget(wtagsel_lbl, 0)
        lo1.addWidget(self.wtagsel, 1)
        self.wtagsel.activated[int].connect(self._check_tag)
        self.wtagsel.editTextChanged['QString'].connect(self._check_tag_text)
        # value editor
        self.valedit = ValueTypeEditor(self)
        lo.addWidget(self.valedit)
        # buttons
        lo.addSpacing(10)
        lo2 = QHBoxLayout()
        lo.addLayout(lo2)
        lo2.setContentsMargins(0, 0, 0, 0)
        lo2.setContentsMargins(5, 5, 5, 5)
        self.wokbtn = QPushButton("OK", self)
        self.wokbtn.setMinimumWidth(128)
        self.wokbtn.clicked.connect(self.accept)
        self.wokbtn.setEnabled(False)
        cancelbtn = QPushButton("Cancel", self)
        cancelbtn.setMinimumWidth(128)
        cancelbtn.clicked.connect(self.reject)
        lo2.addWidget(self.wokbtn)
        lo2.addStretch(1)
        lo2.addWidget(cancelbtn)
        self.setMinimumWidth(384)

    def setTags(self, tagnames):
        self.wtagsel.clear()
        self.wtagsel.addItems(list(tagnames))
        self.wtagsel.addItem("")
        self.wtagsel.setCurrentIndex(len(tagnames))

    def setValue(self, value):
        self.valedit.setValue(value)

    def _check_tag(self, tag):
        self.wokbtn.setEnabled(True)

    def _check_tag_text(self, text):
        self.wokbtn.setEnabled(bool(str(text) != ""))

    def accept(self):
        """When dialog is accepted with a default (bool) tag type,
        check if the user hasn't entered a name=value entry in the tag name field.
        This is a common mistake, and should be treated as a shortcut for setting string tags."""
        if isinstance(self.valedit.getValue(), bool):
            tagval = str(self.wtagsel.currentText()).split("=", 1)
            if len(tagval) > 1:
                #        print tagval
                if QMessageBox.warning(self,
                                       "Set a string tag instead?", """<P>You have included an "=" sign in the tag name.
            Perhaps you actually mean to set tag "%s" to the string value "%s"?</P>""" % tuple(tagval),
                                       QMessageBox.Yes | QMessageBox.No, QMessageBox.Yes) == QMessageBox.No:
                    return
                self.wtagsel.setEditText(tagval[0])
                self.valedit.setValue(tagval[1])
        return QDialog.accept(self)

    def getTag(self):
        return str(self.wtagsel.currentText()), self.valedit.getValue()


class SelectTagsDialog(QDialog):
    def __init__(self, parent, modal=True, flags=Qt.WindowFlags(), caption="Select Tags", ok_button="Select"):
        QDialog.__init__(self, parent, flags)
        self.setModal(modal)
        self.setWindowTitle(caption)
        lo = QVBoxLayout(self)
        lo.setContentsMargins(10, 10, 10, 10)
        lo.setSpacing(5)
        # tag selector
        self.wtagsel = QListWidget(self)
        lo.addWidget(self.wtagsel)
        #    self.wtagsel.setColumnMode(QListBox.FitToWidth)
        self.wtagsel.setSelectionMode(QListWidget.MultiSelection)
        self.wtagsel.itemSelectionChanged.connect(self._check_tag)
        # buttons
        lo.addSpacing(10)
        lo2 = QHBoxLayout()
        lo.addLayout(lo2)
        lo2.setContentsMargins(0, 0, 0, 0)
        lo2.setContentsMargins(5, 5, 5, 5)
        self.wokbtn = QPushButton(ok_button, self)
        self.wokbtn.setMinimumWidth(128)
        self.wokbtn.clicked.connect(self.accept)
        self.wokbtn.setEnabled(False)
        cancelbtn = QPushButton("Cancel", self)
        cancelbtn.setMinimumWidth(128)
        cancelbtn.clicked.connect(self.reject)
        lo2.addWidget(self.wokbtn)
        lo2.addStretch(1)
        lo2.addWidget(cancelbtn)
        self.setMinimumWidth(384)
        self._tagnames = []

    def setTags(self, tagnames):
        self._tagnames = tagnames
        self.wtagsel.clear()
        self.wtagsel.insertItems(0, list(tagnames))

    def _check_tag(self):
        for i in range(len(self._tagnames)):
            if self.wtagsel.item(i).isSelected():
                self.wokbtn.setEnabled(True)
                return
        else:
            self.wokbtn.setEnabled(False)

    def getSelectedTags(self):
        return [tag for i, tag in enumerate(self._tagnames) if self.wtagsel.item(i).isSelected()]


class TDockWidget(QDockWidget):

    def __init__(self, title="", parent=None, flags=Qt.WindowFlags(), bind_widget=None, close_slot=None, toggle_slot=None):
        QDockWidget.__init__(self, title, parent, flags)
        self.installEventFilter(self)
        self.main_win = parent
        # default stlyesheets for title bars
        self.title_stylesheet = "QWidget {background: rgb(68,68,68);}"
        self.button_style = "QPushButton:hover:!pressed {background: grey;}"
        from TigGUI.Images.ControlDialog import ImageControlDialog
        from TigGUI.Plot.SkyModelPlot import ToolDialog
        from TigGUI.Plot.SkyModelPlot import LiveImageZoom
        if bind_widget is not None:
            self.bind_widget = bind_widget
        if bind_widget is not None:
            if isinstance(bind_widget, ToolDialog):
                self.tdock_style = "ToolDialog {border: 1.5px solid rgb(68,68,68);}"
            elif isinstance(bind_widget, ImageControlDialog):
                self.tdock_style = "ImageControlDialog {border: 1.5px solid rgb(68,68,68);}"
        # set default sizes for QDockWidgets
        self.btn_w = 28
        self.btn_h = 28
        self.icon_size = QSize(20, 20)
        self.font_size = 8
        # setup custom title bar for profiles dockable
        self.dock_title_bar = QWidget()
        self.dock_title_bar.setContentsMargins(0, 0, 0, 0)
        self.dock_title_bar.setStyleSheet(self.title_stylesheet)
        self.dock_title_bar.setBaseSize(0, 0)
        self.dock_title_layout = QHBoxLayout()
        self.dock_title_layout.setContentsMargins(0, 0, 0, 0)
        self.dock_title_layout.setSpacing(0)
        self.dock_title_bar.setLayout(self.dock_title_layout)
        # custom close button
        self.close_button = QPushButton()
        self.close_button.setStyleSheet(self.button_style)
        self.close_button.setMaximumWidth(self.btn_w)
        self.close_button.setMaximumHeight(self.btn_h)
        self.close_button.setContentsMargins(0, 0, 0, 0)
        self.close_button.setBaseSize(0, 0)
        self.close_icon = self.dock_title_bar.style().standardIcon(QStyle.SP_TitleBarCloseButton)
        self.close_button.setIcon(self.close_icon)
        self.close_button.setToolTip("Close")
        # custom toggle button
        self.toggle_button = QPushButton()
        self.toggle_button.setStyleSheet(self.button_style)
        self.toggle_button.setMaximumWidth(self.btn_w)
        self.toggle_button.setMaximumHeight(self.btn_h)
        self.toggle_button.setContentsMargins(0, 0, 0, 0)
        self.toggle_button.setBaseSize(0, 0)
        self.toggle_icon = self.dock_title_bar.style().standardIcon(QStyle.SP_TitleBarShadeButton)
        self.toggle_button.setIcon(self.toggle_icon)
        self.toggle_button.setToolTip("Dock/float widget")
        # tigger logo
        self.image0 = pixmaps.tigger_logo.pm()
        self.title_icon = QLabel()
        self.title_icon.setContentsMargins(0, 0, 0, 0)
        self.title_icon.setBaseSize(0, 0)
        self.title_icon.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.title_icon.setScaledContents(True)
        self.title_icon.setPixmap(self.image0)
        self.title_icon.setAlignment(Qt.AlignCenter)
        self.title_icon.setMaximumSize(self.icon_size)
        # set dock widget title
        self.title_font = QFont()
        self.title_font.setBold(True)
        self.title_font.setPointSize(self.font_size)
        if bind_widget is not None:
            if isinstance(bind_widget, ImageControlDialog):
                self.dock_title = QLabel(f"{title}: Control Dialog")
            else:
                self.dock_title = QLabel(title)
        self.dock_title.setFont(self.title_font)
        self.dock_title.setAlignment(Qt.AlignCenter)
        self.dock_title.setContentsMargins(0, 0, 0, 0)
        self.dock_title.setBaseSize(0, 0)
        self.dock_title.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Minimum)
        # add dock widget title items to layout
        self.dock_title_layout.addWidget(self.title_icon)
        self.dock_title_layout.addWidget(self.dock_title)
        self.dock_title_layout.addWidget(self.toggle_button)
        self.dock_title_layout.addWidget(self.close_button)
        # set up profiles as dockable
        self.setStyleSheet(self.tdock_style)
        self.setWidget(bind_widget)
        self.setFeatures(QDockWidget.AllDockWidgetFeatures)
        if bind_widget is not None:
            if isinstance(bind_widget, ToolDialog):
                self.setAllowedAreas(Qt.AllDockWidgetAreas)
            elif isinstance(bind_widget, ImageControlDialog):
                self.setAllowedAreas(Qt.RightDockWidgetArea | Qt.LeftDockWidgetArea)
        self.setTitleBarWidget(self.dock_title_bar)
        self.setFloating(False)
        # get current sizeHints()
        if bind_widget is not None:
            self.setBaseSize(bind_widget.sizeHint())
            if isinstance(bind_widget, LiveImageZoom):
                bind_widget.livezoom_resize_signal.connect(self._resizeDockWidget)
        if close_slot is not None:
            self.close_button.clicked.connect(close_slot)
            if isinstance(bind_widget, ImageControlDialog):
                bind_widget.whide.clicked.connect(close_slot)
        if toggle_slot is not None:
            self.toggle_button.clicked.connect(toggle_slot)

    def _resizeDockWidget(self, qsize):
        # live zoom signal slot to resize dockwidget and dock areas
        self.setMinimumSize(qsize)
        self.main_win.resizeDocks([self], [qsize.width()], Qt.Horizontal)
        self.main_win.resizeDocks([self], [qsize.height()], Qt.Vertical)

    # hack to stop QDockWidget responding to drag events for undocking - work around for Qt bug
    def eventFilter(self, source, event):
        # event seq 2, 5, 3 - mouse press, mouse move, mouse release
        if event.type() == QEvent.MouseButtonPress:
            label = self.childAt(event.pos())
            if not label:
                return super(TDockWidget, self).eventFilter(source, event)
            if isinstance(label, QLabel):
                if not self.isFloating():
                    fake_mouse_event = QMouseEvent(QEvent.MouseButtonRelease, event.pos(), event.button(), event.buttons(), event.modifiers())
                    super(TDockWidget, self).event(fake_mouse_event)
                    return True
        return super(TDockWidget, self).eventFilter(source, event)


class TigToolTip(QLabel):
    """Custom QToolTip type widget based on a QLabel for plot information output."""
    def __init__(self):
        QLabel.__init__(self)
        self.installEventFilter(self)
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)
        self.setAttribute(Qt.WA_QuitOnClose)
        self.setStyleSheet("QLabel {background-color: white; color: black;}")
        self.setTextFormat(Qt.RichText)
        self.setWordWrap(True)
        self._qtimer = QTimer()
        self._qtimer.timeout.connect(self.hideText)

    def showText(self, location, text, timeout=7000):
        if self._qtimer.isActive():
            self._qtimer.start(timeout)
        self.setText(text)
        text_size = self.fontMetrics().boundingRect(self.text())
        # TODO - find a better way for the below sizes
        if text_size.width() > 900:
            max_w = 700
            max_h = text_size.height() * 4
        else:
            max_w = 900
            max_h = text_size.height()
        self.setGeometry(location.x(), location.y(), max_w, max_h)
        self.show()
        self._qtimer.start(timeout)

    def hideText(self):
        self._qtimer.stop()
        self.close()
        # available on Ubuntu by default
        # disabling for now issue #163
        # os.system('notify-send "Tigger" "Information copied to clipboard"')

    def eventFilter(self, source, event):
        # event.type() 25 == QEvent.WindowDeactivate.
        # In this context, TigToolTip is the top most window and when application has been changed in terms of state,
        # for example to another application, the TigToolTip needs to be closed, otherwise it will remain on the screen.
        if event.type() == QEvent.WindowDeactivate:
            if self._qtimer.isActive():
                self._qtimer.stop()
            self.close()
        return super(TigToolTip, self).eventFilter(source, event)



# Copyright (C) 2002-2022
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import math
from builtins import chr

from PyQt5.Qt import QWidget, QHBoxLayout, QComboBox, QLabel, QToolButton, QVBoxLayout, \
    QPushButton, Qt, QTreeWidgetItem, QAbstractItemView, QHeaderView, QTreeWidget, QAction, QEvent, QSize, \
    QSizePolicy, QTableWidget, QTableWidgetItem, QItemSelectionRange, QItemSelection, QFontMetrics, QFont, \
    QApplication, QItemSelectionModel
from Tigger.Models import ModelClasses, PlotStyles
from Tigger.Models.SkyModel import SkyModel

import TigGUI.kitties.utils
import TigGUI.kitties.widgets
from TigGUI.kitties.utils import PersistentCurrier
from TigGUI.kitties.widgets import BusyIndicator

QString = str

_verbosity = TigGUI.kitties.utils.verbosity(name="tw")
dprint = _verbosity.dprint
dprintf = _verbosity.dprintf

ViewColumns = ["name", "RA", "RA err", "Dec", "Dec err", "r", "type",
               "Iapp", "I", "I err", "Q", "Q err", "U", "U err", "V", "V err", "RM", "RM err", "spi", "spi err",
               "shape", "shape err", "tags"]

# ugly, but better than the global() hack this was before.
ColumnName = 0
ColumnRa = 1
ColumnRa_err = 2
ColumnDec = 3
ColumnDec_err = 4
ColumnR = 5
ColumnType = 6
ColumnIapp = 7
ColumnI = 8
ColumnI_err = 9
ColumnQ = 10
ColumnQ_err = 11
ColumnU = 12
ColumnU_err = 13
ColumnV = 14
ColumnV_err = 15
ColumnRm = 16
ColumnRm_err = 17
ColumnSpi = 18
ColumnSpi_err = 19
ColumnShape = 20
ColumnShape_err = 21
ColumnTags = 22

NumColumns = len(ViewColumns)

DEG = math.pi / 180

# Qt-4.6 and up (PyQt 4.7 and up) has very slow QTreeWidgetItem updates, determine version here
from PyQt5 import QtCore

_SLOW_QTREEWIDGETITEM = QtCore.PYQT_VERSION_STR >= '4.7'


class SkyModelTreeWidget(TigGUI.kitties.widgets.ClickableTreeWidget):
    """This implements a QTreeWidget for sky models"""

    def __init__(self, *args):
        TigGUI.kitties.widgets.ClickableTreeWidget.__init__(self, *args)
        self._currier = PersistentCurrier()
        self.model = None
        # insert columns
        self.setHeaderLabels(ViewColumns)
        self.headerItem().setText(ColumnIapp, "I(app)")
        self.header().setSectionsMovable(False)
        self.header().setSectionsClickable(True)
        self.setSortingEnabled(True)
        self.setRootIsDecorated(False)
        self.setEditTriggers(QAbstractItemView.AllEditTriggers)
        self.setMouseTracking(True)
        # set column width modes
        self.header().setStretchLastSection(False)
        self.header().setSectionResizeMode(QHeaderView.ResizeToContents)
        ## self.setTextAlignment(ColumnR,Qt.AlignRight)
        ## self.setTextAlignment(ColumnType,Qt.AlignHCenter)
        # _column_enabled[i] is True if column is available in the model.
        # _column_show[i] is True if column is currently being shown (via a view control)
        self._column_enabled = [True] * NumColumns
        self._column_shown = [True] * NumColumns
        # other listview init
        self.header().show()
        self.setSelectionMode(QTreeWidget.ExtendedSelection)
        self.setAllColumnsShowFocus(True)
        ## self.setShowToolTips(True)
        self._updating_selection = False
        self.setRootIsDecorated(False)
        # connect signals to track selected sources
        self.itemSelectionChanged.connect(self._selectionChanged)
        self.itemEntered[QTreeWidgetItem, int].connect(self._itemHighlighted)
        # add "View" controls for different column categories
        self._column_views = []
        self._column_widths = {}
        self.addColumnCategory("Position", [ColumnRa, ColumnDec])
        self.addColumnCategory("Position errors", [ColumnRa_err, ColumnDec_err], False)
        self.addColumnCategory("Type", [ColumnType])
        self.addColumnCategory("Flux", [ColumnIapp, ColumnI])
        self.addColumnCategory("Flux errors", [ColumnI_err], False)
        self.addColumnCategory("Polarization", [ColumnQ, ColumnU, ColumnV, ColumnRm])
        self.addColumnCategory("Polarization errors", [ColumnQ_err, ColumnU_err, ColumnV_err, ColumnRm_err], False)
        self.addColumnCategory("Spectrum", [ColumnSpi])
        self.addColumnCategory("Spectrum errors", [ColumnSpi_err], False)
        self.addColumnCategory("Shape", [ColumnShape])
        self.addColumnCategory("Shape errors", [ColumnShape_err], False)
        self.addColumnCategory("Tags", [ColumnTags])

    def _showColumn(self, col, show=True):
        """Shows or hides the specified column.
        (When hiding, saves width of column to internal array so that it can be restored properly.)"""
        hdr = self.header()
        hdr.setSectionHidden(col, not show)
        if show:
            if not hdr.sectionSize(col):
                hdr.resizeSection(col, self._column_widths[col])
                hdr.setResizeMode(col, QHeaderView.ResizeToContents)
        else:
            if hdr.sectionSize(col):
                self._column_widths[col] = hdr.sectionSize(col)

    def _enableColumn(self, column, enable=True):
        busy = BusyIndicator()
        self._column_enabled[column] = enable
        self._showColumn(column, enable and self._column_shown[column])
        busy.reset_cursor()

    def _showColumnCategory(self, columns, show):
        busy = BusyIndicator()
        for col in columns:
            self._column_shown[col] = show
            self._showColumn(col, self._column_enabled[col] and show)
        busy.reset_cursor()

    def _selectionChanged(self):
        if self._updating_selection:
            return
        for item in self.iterator():
            item._src.select(item.isSelected())
        self.model.emitSelection(origin=self)

    def _itemHighlighted(self, item, col):
        dprint(3, "highlighting", item._src.name)
        self.model.setCurrentSource(item._src, origin=self)

    def viewportEvent(self, event):
        if event.type() in (QEvent.Leave, QEvent.FocusOut) and self.model:
            self.model.setCurrentSource(None, origin=self)
        return QTreeWidget.viewportEvent(self, event)

    def addColumnCategory(self, name, columns, visible=True):
        qa = QAction(name, self)
        qa.setCheckable(True)
        qa.setChecked(visible)
        if not visible:
            self._showColumnCategory(columns, False)
        qa.toggled[bool].connect(self._currier.curry(self._showColumnCategory, columns))
        self._column_views.append((name, qa, columns))

    def clear(self):
        TigGUI.kitties.widgets.ClickableTreeWidget.clear(self)
        self.model = None
        self._itemdict = {}

    def setModel(self, model):
        self.model = model
        self._refreshModel(SkyModel.UpdateAll)
        self.model.connect("changeCurrentSource", self._updateCurrentSource)
        self.model.connect("changeGroupingVisibility", self.changeGroupingVisibility)
        self.model.connect("selected", self._updateModelSelection)
        self.model.connect("updated", self._refreshModel)

    def _refreshModel(self, what=SkyModel.UpdateAll, origin=None):
        if origin is self or not what & (SkyModel.UpdateSourceList | SkyModel.UpdateSourceContent):
            return
        # if only selection was changed, take shortcut
        if what & SkyModel.UpdateSelectionOnly:
            dprint(2, "model update -- selection only")
            return self._refreshSelectedItems(origin)
        busy = BusyIndicator()
        # else repopulate widget completely
        dprint(2, "model update -- complete")
        TigGUI.kitties.widgets.ClickableTreeWidget.clear(self)
        dprint(2, "creating model items")
        items = [SkyModelTreeWidgetItem(src) for src in self.model.sources]
        self._itemdict = dict(list(zip([src.name for src in self.model.sources], items)))
        dprint(2, "adding to tree widget")
        self.addTopLevelItems(items)
        self.header().updateGeometry()
        # show/hide columns based on tag availability
        self._enableColumn(ColumnIapp, 'Iapp' in self.model.tagnames)
        self._enableColumn(ColumnR, 'r' in self.model.tagnames)
        dprint(2, "re-sorting")
        self.sortItems(('Iapp' in self.model.tagnames and ColumnIapp) or ColumnI, Qt.DescendingOrder)
        busy.reset_cursor()

    def addColumnViewActionsTo(self, menu):
        for name, qa, columns in self._column_views:
            menu.addAction(qa)

    def _updateCurrentSource(self, src, src0=None, origin=None):
        # if origin is self:
        # return
        # dehighlight old item
        item = src0 and self._itemdict.get(src0.name)
        if item:
            item.setHighlighted(False)
        # scroll to new item, if found
        item = src and self._itemdict.get(src.name)
        if item:
            item.setHighlighted(True, origin is not self)
            if origin is not self:
                self.scrollToItem(item)

    def _updateModelSelection(self, nsel, origin=None):
        """This is called when some other widget (origin!=self) changes the set of selected model sources"""
        if origin is self:
            return
        self._updating_selection = True
        ## this is very slow because of setSelected()
        #    for item in self.iterator():
        #     item.setSelected(item._src.selected)
        selection = QItemSelection()
        for item in self.iterator():
            if item._src.selected:
                selection.append(
                    QItemSelectionRange(self.indexFromItem(item, 0), self.indexFromItem(item, self.columnCount() - 1)))
        self.selectionModel().select(selection, QItemSelectionModel.ClearAndSelect)
        self.changeGroupingVisibility(None, origin=origin)
        self._updating_selection = False

    def _refreshSelectedItems(self, origin=None):
        busy = BusyIndicator()
        dprint(3, "refreshing selected items")
        for item in self.iterator():
            if item.isSelected():
                dprint(4, "resetting item", item._src.name)
                item.setSource(item._src)
        dprint(3, "refreshing selected items done")
        busy.reset_cursor()

    def changeGroupingVisibility(self, group, origin=None):
        if origin is self:
            return
        for item in self.iterator():
            # collect show_list values from groupings to which this source belongs (default group excepted)
            show = [group.style.show_list for group in self.model.groupings if
                    group is not self.model.defgroup and group.func(item._src)]
            # if at least one group is showing explicitly, show
            # else if at least one group is hiding explicitly, hide
            # else use default setting
            if max(show) == PlotStyles.ShowAlways:
                visible = True
            elif min(show) == PlotStyles.ShowNot:
                visible = False
            else:
                visible = bool(self.model.defgroup.style.show_list)
            # set visibility accordingly
            item.setHidden(not visible)

    TagsWithOwnColumn = set(["Iapp", "r"])


class SkyModelTreeWidgetItem(QTreeWidgetItem):
    _fonts = None
    _fontmetrics = None

    @staticmethod
    def _initFonts():
        """Initializes fonts on fitrst call"""
        if SkyModelTreeWidgetItem._fonts is None:
            stdfont = QApplication.font()
            boldfont = QFont(stdfont)
            boldfont.setBold(True)
            SkyModelTreeWidgetItem._fonts = [stdfont, boldfont]
            SkyModelTreeWidgetItem._fontmetrics = QFontMetrics(boldfont)

    def __init__(self, src, *args):
        QTreeWidgetItem.__init__(self, *args)
        self._src = src
        # fonts
        self._initFonts()
        # array of actual (i.e. numeric) column values
        self._values = [0.0] * NumColumns
        # set text alignment
        for icol in range(NumColumns):
            self.setTextAlignment(icol, Qt.AlignLeft)
        self.setTextAlignment(ColumnR, Qt.AlignRight)
        self.setTextAlignment(ColumnType, Qt.AlignHCenter)
        # setup source
        self._highlighted = self._highlighted_visual = False
        self.setSource(src)

    def setHighlighted(self, highlighted=True, visual=False):
        #    global _SLOW_QTREEWIDGETITEM
        #    if 1: # not _SLOW_QTREEWIDGETITEM:
        visual = True
        dprint(3, self._src.name, "highlighted", highlighted, visual)
        if highlighted != self._highlighted:
            #      brush = QApplication.palette().alternateBase() if highlighted else QApplication.palette().base()
            #      for col in range(self.columnCount()):
            #        self.setBackground(col,brush)
            if highlighted and visual:
                self.setFont(0, self._fonts[1])
            elif not highlighted and self._highlighted_visual:
                self.setFont(0, self._fonts[0])
            self._highlighted = highlighted
            self._highlighted_visual = visual

    @staticmethod
    def _angErrToStr(value):
        """helper method: converts angular error to string representation in deg or arcmin or arcsec"""
        arcsec = (value / DEG) * 3600
        if arcsec < 60:
            return chr(0xB1) + "%.2g\"" % arcsec
        elif arcsec < 3600:
            return chr(0xB1) + "%.2f'" % (arcsec * 60)
        else:
            return chr(0xB1) + "%.2f%s" % (arcsec * 3600, chr(0xB0))

    def setSource(self, src):
        # name
        dprint(3, "setSource 1", src.name)
        self.setColumn(ColumnName, src.name)
        self.setSizeHint(0, QSize(self._fontmetrics.width("x" + src.name), 0))
        # coordinates
        self.setColumn(ColumnRa, src.pos.ra, "%2dh%02dm%05.2fs" % src.pos.ra_hms())
        self.setColumn(ColumnDec, src.pos.dec, ("%s%2d" + chr(0xB0) + "%02d'%05.2f\"") %
                       src.pos.dec_sdms())
        if src.pos.ra_err is not None:
            self.setColumn(ColumnRa_err, src.pos.ra_err, self._angErrToStr(src.pos.ra_err))
        if src.pos.dec_err is not None:
            self.setColumn(ColumnDec_err, src.pos.dec_err, self._angErrToStr(src.pos.dec_err))
        if hasattr(src, 'r'):
            self.setColumn(ColumnR, src.r, "%.1f'" % (src.r * 180 * 60 / math.pi))
        # type
        self.setColumn(ColumnType, src.typecode)
        # flux
        if hasattr(src, 'Iapp'):
            self.setColumn(ColumnIapp, src.Iapp, "%.3g" % src.Iapp)
        for stokes in "IQUV":
            stk = getattr(src.flux, stokes, None)
            stk_err = getattr(src.flux, stokes + "_err", None)
            if stk is not None:
                self.setColumn(globals()['Column' + stokes], stk, "%.3g" % stk)
            if stk_err is not None:
                self.setColumn(globals()['Column' + stokes + "_err"], stk_err, chr(0xB1) + "%.2g" % stk_err)
        if hasattr(src.flux, 'rm'):
            self.setColumn(ColumnRm, src.flux.rm, "%.2f" % src.flux.rm)
            if hasattr(src.flux, 'rm_err'):
                self.setColumn(ColumnRm_err, src.flux.rm_err, chr(0xB1) + "%.2f" % src.flux.rm)
        # spi
        if isinstance(src.spectrum, ModelClasses.SpectralIndex):
            spi = getattr(src.spectrum, 'spi', 0)
            if not isinstance(spi, (list, tuple)):
                spi = [spi]
            spi = ",".join(["%.2f" % x for x in spi])
            self.setColumn(ColumnSpi, src.spectrum.spi, spi)
            spierr = getattr(src.spectrum, 'spi_err', None)
            if spierr is not None:
                if not isinstance(spierr, (list, tuple)):
                    spierr = [spierr]
                spierr = ",".join(["%.2f" % x for x in spierr])
                self.setColumn(ColumnSpi_err, src.spectrum.spi_err, chr(0xB1) + spierr)
        # shape
        shape = getattr(src, 'shape', None)
        if isinstance(shape, ModelClasses.ModelItem):
            shapeval = shape.getShape()
            shapestr = shape.strDesc(delimiters=('"', chr(0xD7), chr(0x21BA), chr(0xB0)))
            self.setColumn(ColumnShape, shapeval, shapestr)
            errval = shape.getShapeErr()
            if errval:
                errstr = shape.strDescErr(delimiters=('"', chr(0xD7), chr(0x21BA), chr(0xB0)))
                self.setColumn(ColumnShape_err, errval, chr(0xB1) + errstr)
        dprint(3, "setSource 3", src.name)
        # Tags. Tags are all extra attributes that do not have a dedicated column (i.e. not Iapp or r), and do not start
        # with "_" (which is reserved for internal attributes)

        ## the complexity below seems entirely unnecessary, since sorting the tag strings automatically puts "_" first,
        ## "-" second, and alphabet afterwards

        # truetags = []
        # falsetags = []
        # othertags = []
        # for attr,val in src.getExtraAttributes():
        # if attr[0] != "_" and attr not in SkyModelTreeWidget.TagsWithOwnColumn:
        # if val is False:
        # falsetags.append("-"+attr)
        # elif val is True:
        # truetags.append("+"+attr)
        # else:
        # othertags.append("%s=%s"%(attr,str(val)))
        # for tags in truetags,falsetags,othertags:
        # tags.sort()
        # self.setColumn(ColumnTags,tags," ".join(truetags+falsetags+othertags))

        # so instead:
        tags = ["+" + attr if val is True else "-" + attr if val is False else "%s=%s" % (attr, str(val))
                for attr, val in src.getExtraAttributes()
                if attr[0] != "_" and attr not in SkyModelTreeWidget.TagsWithOwnColumn]
        tagstr = " ".join(sorted(tags))
        dprint(3, "setSource 4", src.name)
        self.setColumn(ColumnTags, tags, tagstr)
        dprint(3, "setSource 5", src.name)
        dprint(3, "setSource done", src.name)

    def setColumn(self, icol, value, text=None):
        """helper function to set the value of a column"""
        if text is None:
            text = str(value)
        self.setText(icol, text)
        self._values[icol] = value

    def __lt__(self, other):
        icol = self.treeWidget().sortColumn()
        if icol is not None:
            if isinstance(other, SkyModelTreeWidgetItem):
                if self._values[icol] is not None and other._values[icol] is not None:
                    if isinstance(self._values[icol], type(other._values[icol])):
                        return self._values[icol] < other._values[icol]
                    else:
                        return False
                else:
                    return False
            else:
                if self._text(icol) is not None and other.text(icol) is not None:
                    if isinstance(self._text(icol), type(other.text(icol))):
                        return self.text(icol) < other.text(icol)
                    else:
                        return False
                else:
                    return False
        else:
            return False

    def __ge__(self, other):
        return other < self


class ModelGroupsTable(QWidget):
    EditableAttrs = [attr for attr in PlotStyles.StyleAttributes if attr in PlotStyles.StyleAttributeOptions]
    ColList = 3
    ColPlot = 4
    ColApply = 5
    AttrByCol = dict([(i + 6, attr) for i, attr in enumerate(EditableAttrs)])

    def __init__(self, parent, *args):
        QWidget.__init__(self, parent, *args)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        lo = QVBoxLayout(self)
        lo.setContentsMargins(0, 0, 0, 0)
        lo1 = QHBoxLayout()
        lo.addLayout(lo1)
        lo1.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel(QString("<nobr><b>Source groupings:</b></nobr>"), self)
        lo1.addWidget(lbl, 0)
        lo1.addStretch(1)
        # add show/hide button
        self._showattrbtn = QPushButton(self)
        self._showattrbtn.setMinimumWidth(256)
        lo1.addWidget(self._showattrbtn, 0)
        lo1.addStretch()
        self._showattrbtn.clicked.connect(self._togglePlotControlsVisibility)
        # add table
        self.table = QTableWidget(self)
        lo.addWidget(self.table)
        self.table.cellChanged[int, int].connect(self._valueChanged)
        self.table.setSelectionMode(QTableWidget.NoSelection)
        # setup basic columns
        self.table.setColumnCount(6 + len(self.EditableAttrs))
        for i, label in enumerate(("grouping", "total", "selection", "list", "plot", "style")):
            self.table.setHorizontalHeaderItem(i, QTableWidgetItem(label))
        self.table.horizontalHeader().setSectionHidden(self.ColApply, True)
        # setup columns for editable grouping attributes
        for i, attr in self.AttrByCol.items():
            self.table.setHorizontalHeaderItem(i, QTableWidgetItem(PlotStyles.StyleAttributeLabels[attr]))
            self.table.horizontalHeader().setSectionHidden(i, True)
        self.table.verticalHeader().hide()
        # other internal init
        self._attrs_shown = False
        self._togglePlotControlsVisibility()
        self.model = None
        self._setting_model = False
        self._currier = PersistentCurrier()
        # row of 'selected' grouping
        self._irow_selgroup = 0

    def clear(self):
        self.table.setRowCount(0)
        self.model = None

    # setup mappings from the group.show_plot attribute to check state
    ShowAttrToCheckState = {PlotStyles.ShowNot: Qt.Unchecked,
                            PlotStyles.ShowDefault: Qt.PartiallyChecked,
                            PlotStyles.ShowAlways: Qt.Checked}
    CheckStateToShowAttr = dict([(val, key) for key, val in ShowAttrToCheckState.items()])

    def _makeCheckItem(self, name, group, attr):
        item = QTableWidgetItem(name)
        if group is self.model.defgroup:
            item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsUserCheckable)
            item.setCheckState(Qt.Checked if getattr(group.style, attr) else Qt.Unchecked)
        else:
            item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsUserCheckable | Qt.ItemIsTristate)
            item.setCheckState(self.ShowAttrToCheckState[getattr(group.style, attr)])
        return item

    def _updateModel(self, what=SkyModel.UpdateAll, origin=None):
        if origin is self or not what & (SkyModel.UpdateTags | SkyModel.UpdateGroupStyle):
            return
        model = self.model
        self._setting_model = True  # to ignore cellChanged() signals (in valueChanged())
        # _item_cb is a dict (with row,col keys) containing the widgets (CheckBoxes ComboBoxes) per each cell
        self._item_cb = {}
        # lists of "list" and "plot" checkboxes per each grouping (excepting the default grouping); each entry is an (row,col,item) tuple.
        # used as argument to self._showControls()
        self._list_controls = []
        self._plot_controls = []
        # list of selection callbacks (to which signals are connected)
        self._callbacks = []
        # set requisite number of rows,and start filling
        self.table.setRowCount(len(model.groupings))
        for irow, group in enumerate(model.groupings):
            self.table.setItem(irow, 0, QTableWidgetItem(group.name))
            if group is model.selgroup:
                self._irow_selgroup = irow
            # total # source in group: skip for "current"
            if group is not model.curgroup:
                self.table.setItem(irow, 1, QTableWidgetItem(str(group.total)))
            # selection controls: skip for current and selection
            if group not in (model.curgroup, model.selgroup):
                btns = QWidget()
                lo = QHBoxLayout(btns)
                lo.setContentsMargins(0, 0, 0, 0)
                lo.setSpacing(0)
                # make selector buttons (depending on which group we're in)
                if group is model.defgroup:
                    Buttons = (
                        ("+", lambda src, grp=group: True, "select all sources"),
                        ("-", lambda src, grp=group: False, "unselect all sources"))
                else:
                    Buttons = (
                        ("=", lambda src, grp=group: grp.func(src), "select only this grouping"),
                        ("+", lambda src, grp=group: src.selected or grp.func(src), "add grouping to selection"),
                        ("-", lambda src, grp=group: src.selected and not grp.func(src),
                         "remove grouping from selection"),
                        ("&&", lambda src, grp=group: src.selected and grp.func(src),
                         "intersect selection with grouping"))
                lo.addStretch(1)
                for label, predicate, tooltip in Buttons:
                    btn = QToolButton(btns)
                    btn.setText(label)
                    btn.setMinimumWidth(24)
                    btn.setMaximumWidth(24)
                    btn.setToolTip(tooltip)
                    lo.addWidget(btn)
                    # add callback
                    btn.clicked.connect(self._currier.curry(self.selectSources, predicate))
                lo.addStretch(1)
                self.table.setCellWidget(irow, 2, btns)
            # "list" checkbox (not for current and selected groupings: these are always listed)
            if group not in (model.curgroup, model.selgroup):
                item = self._makeCheckItem("", group, "show_list")
                self.table.setItem(irow, self.ColList, item)
                item.setToolTip("""<P>If checked, sources in this grouping will be listed in the source table. If un-checked, sources will be
            excluded from the table. If partially checked, then the default list/no list setting of "all sources" will be in effect.
            </P>""")
            # "plot" checkbox (not for the current grouping, since that's always plotted)
            if group is not model.curgroup:
                item = self._makeCheckItem("", group, "show_plot")
                self.table.setItem(irow, self.ColPlot, item)
                item.setToolTip("""<P>If checked, sources in this grouping will be included in the plot. If un-checked, sources will be
            excluded from the plot. If partially checked, then the default plot/no plot setting of "all sources" will be in effect.
            </P>""")
            # custom style control
            # for default, current and selected, this is just a text label
            if group is model.defgroup:
                item = QTableWidgetItem("default:")
                item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
                item.setToolTip(
                    """<P>This is the default plot style used for all sources for which a custom grouping style is not selected.</P>""")
                self.table.setItem(irow, self.ColApply, item)
            elif group is model.curgroup:
                item = QTableWidgetItem("")
                item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
                item.setToolTip("""<P>This is the plot style used for the highlighted source, if any.</P>""")
                self.table.setItem(irow, self.ColApply, item)
            elif group is model.selgroup:
                item = QTableWidgetItem("")
                item.setTextAlignment(Qt.AlignRight | Qt.AlignVCenter)
                item.setToolTip("""<P>This is the plot style used for the currently selected sources.</P>""")
                self.table.setItem(irow, self.ColApply, item)
            # for the rest, a combobox with custom priorities
            else:
                cb = QComboBox()
                cb.addItems(["default"] + ["custom %d" % p for p in range(1, 10)])
                index = max(0, min(group.style.apply, 9))
                #        dprint(0,group.name,"apply",index)
                cb.setCurrentIndex(index)
                cb.activated[int].connect(self._currier.xcurry(self._valueChanged, (irow, self.ColApply)))
                self.table.setCellWidget(irow, self.ColApply, cb)
                cb.setToolTip("""<P>This controls whether sources within this group are plotted with a customized
            plot style. Customized styles have numeric priority; if a source belongs to multiple groups, then
            the style with the lowest priority takes precedence.<P>""")
            # attribute comboboxes
            for icol, attr in self.AttrByCol.items():
                # get list of options for this style attribute. If dealing with first grouping (i==0), which is
                # the "all sources" grouping, then remove the "default" option (which is always first in the list)
                options = PlotStyles.StyleAttributeOptions[attr]
                if irow == 0:
                    options = options[1:]
                # make combobox
                cb = QComboBox()
                cb.addItems(list(map(str, options)))
                # the "label" option is also editable
                if attr == "label":
                    cb.setEditable(True)
                try:
                    index = options.index(getattr(group.style, attr))
                    cb.setCurrentIndex(index)
                except ValueError:
                    cb.setEditText(str(getattr(group.style, attr)))
                slot = self._currier.xcurry(self._valueChanged, (irow, icol))
                cb.activated[int].connect(slot)
                cb.editTextChanged['QString'].connect(slot)
                cb.setEnabled(group is model.defgroup or group.style.apply)
                self.table.setCellWidget(irow, icol, cb)
                label = attr
                if irow:
                    cb.setToolTip("""<P>This is the %s used to plot sources in this group, when a "custom" style for the group
          is enabled via the style control.<P>""" % label)
                else:
                    cb.setToolTip(
                        "<P>This is the default %s used for all sources for which a custom style is not specified below.<P>" % label)
        self.table.resizeColumnsToContents()
        # re-enable processing of cellChanged() signals
        self._setting_model = False

    def setModel(self, model):
        self.model = model
        self.model.connect("updated", self._updateModel)
        self.model.connect("selected", self.updateModelSelection)
        self._updateModel(SkyModel.UpdateAll)

    def _valueChanged(self, row, col):
        """Called when a cell has been edited"""
        if self._setting_model:
            return
        group = self.model.groupings[row]
        item = self.table.item(row, col)
        if col == self.ColList:
            if group is not self.model.defgroup:
                # tri-state items go from unchecked to checked when user clicks them. Make them partially checked instead.
                if group.style.show_list == PlotStyles.ShowNot and item.checkState() == Qt.Checked:
                    item.setCheckState(Qt.PartiallyChecked)
            group.style.show_list = self.CheckStateToShowAttr[item.checkState()]
            self.model.emitChangeGroupingVisibility(group, origin=self)
            return
        elif col == self.ColPlot:
            if group is not self.model.defgroup:
                # tri-state items go from unchecked to checked by default. Make them partially checked instead.
                if group.style.show_plot == PlotStyles.ShowNot and item.checkState() == Qt.Checked:
                    item.setCheckState(Qt.PartiallyChecked)
            group.style.show_plot = self.CheckStateToShowAttr[item.checkState()]
        elif col == self.ColApply:
            group.style.apply = self.table.cellWidget(row, col).currentIndex()
            # enable/disable editable cells
            for j in list(self.AttrByCol.keys()):
                item1 = self.table.item(row, j)
                if item1:
                    fl = item1.flags() & ~Qt.ItemIsEnabled
                    if group.style.apply:
                        fl |= Qt.ItemIsEnabled
                    item1.setFlags(fl)
                cw = self.table.cellWidget(row, j)
                cw and cw.setEnabled(group.style.apply)
        elif col in self.AttrByCol:
            cb = self.table.cellWidget(row, col)
            txt = str(cb.currentText())
            attr = self.AttrByCol[col]
            if txt == "default":
                setattr(group.style, attr, PlotStyles.DefaultValue)
            else:
                setattr(group.style, attr, PlotStyles.StyleAttributeTypes.get(attr, str)(txt))
        # all other columns: return so we don't emit a signal
        else:
            return
        # in all cases emit a signal
        self.model.emitChangeGroupingStyle(group, origin=self)

    def selectSources(self, predicate, curry=False):
        """Selects sources according to predicate(src)"""
        busy = BusyIndicator()
        for src in self.model.sources:
            src.selected = predicate(src)
        self.model.emitSelection(origin=self)
        busy.reset_cursor()

    def updateModelSelection(self, nsel, origin=None):
        """This is called when some other widget changes the set of selected model sources"""
        self.table.clearSelection()
        if self.model:
            self.table.item(self._irow_selgroup, 1).setText(str(nsel))

    def _togglePlotControlsVisibility(self):
        if self._attrs_shown:
            self._attrs_shown = False
            self.table.hideColumn(self.ColApply)
            for col in self.AttrByCol.keys():
                self.table.hideColumn(col)
            self._showattrbtn.setText("Show plot styles >>")
        else:
            self._attrs_shown = True
            self.table.showColumn(self.ColApply)
            for col in self.AttrByCol.keys():
                self.table.showColumn(col)
            self._showattrbtn.setText("<< Hide plot styles")



#!/usr/bin/env python3

from __future__ import print_function
from setuptools import setup, find_packages
from pathlib import Path

__version__ = "1.6.1.1"

# PyQt (5.15.x) has not been added here are it needs to be installed via apt-get instead to support Qwt.
# requirements below do not have versions for upstream packaging processes, but tested and compatible versions are given
requirements = [
    'numpy',  # tested with version >=1.19.4 (<= 1.22.3)
    'scipy',  # tested with versions 1.5.2 for Python 3.6 and >=1.6.2 (<= 1.8.0)
    'astlib',  # tested with version 0.11.6 and 0.11.7
    'astropy',  # tested with 4.1, 4.2 and 5.0.4
    'astro_tigger_lsm==1.7.1',  # PyQt5 version of astro-tigger-lsm
    'configparser',  # tested with version 5.0.1 and 5.2.0
]

scripts = [
    'TigGUI/tigger',
]

package_data = {'TigGUI': [
        'icons/*.png',
        'tigger.conf',
    ]
}


setup(
    name="astro-tigger",
    version=__version__,
    packages=find_packages(),
    scripts=scripts,
    package_data=package_data,
    description="yet another FITS image viewer",
    author="Oleg Smirnov",
    author_email="osmirnov@gmail.com",
    url="https://github.com/ratt-ru/tigger",
    python_requires='>=3.6',
    install_requires=requirements,
    data_files=[
        (f"{Path.home()}/.local/share/applications", ['desktop/tigger.desktop']),
        (f"{Path.home()}/.local/share/icons", ['TigGUI/icons/tigger_logo.png']),
    ],
)




# Copyright (C) 2002-2022
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import math
from PyQt5.QtWidgets import *


import os.path
from PyQt5.Qt import QObject, QHBoxLayout, pyqtSignal, QDialog, QVBoxLayout, \
    QPushButton, Qt, QCheckBox, QMessageBox, QErrorMessage

from TigGUI.kitties.widgets import BusyIndicator
from TigGUI.Widgets import FileSelector

DEG = math.pi / 180


class ExportKarmaDialog(QDialog):
    def __init__(self, parent, modal=True, flags=Qt.WindowFlags()):
        QDialog.__init__(self, parent, flags)
        self.model = None
        self.setModal(modal)
        self.setWindowTitle("Export Karma annotations")
        lo = QVBoxLayout(self)
        lo.setContentsMargins(10, 10, 10, 10)
        lo.setSpacing(5)
        # file selector
        self.wfile = FileSelector(self, label="Filename:", dialog_label="Karma annotations filename",
                                  default_suffix="ann", file_types="Karma annotations (*.ann)")
        lo.addWidget(self.wfile)
        # selected sources checkbox
        self.wsel = QCheckBox("selected sources only", self)
        lo.addWidget(self.wsel)
        # OK/cancel buttons
        lo.addSpacing(10)
        lo2 = QHBoxLayout()
        lo.addLayout(lo2)
        lo2.setContentsMargins(5, 5, 5, 5)
        self.wokbtn = QPushButton("OK", self)
        self.wokbtn.setMinimumWidth(128)
        self.wokbtn.clicked.connect(self.accept)
        self.wokbtn.setEnabled(False)
        cancelbtn = QPushButton("Cancel", self)
        cancelbtn.setMinimumWidth(128)
        cancelbtn.clicked.connect(self.reject)
        lo2.addWidget(self.wokbtn)
        lo2.addStretch(1)
        lo2.addWidget(cancelbtn)
        self.setMinimumWidth(384)
        # signals
        self.wfile.valid.connect(self.wokbtn.setEnabled)
        # internal state
        self.qerrmsg = QErrorMessage(self)
        self._model_filename = None

    def setModel(self, model):
        self.model = model
        # set the default annotations filename, whenever a new model filename is set
        filename = self.model.filename()
        if filename and filename != self._model_filename:
            self._model_filename = filename
            self.wfile.setFilename(os.path.splitext(filename)[0] + ".ann")

    def accept(self):
        """Tries to export annotations, and closes the dialog if successful."""
        try:
            filename = self.wfile.filename()
            if os.path.exists(filename) and QMessageBox.question(self, "Exporting Karma annotations",
                                                                 "<P>Overwrite the file %s?</P>" % filename,
                                                                 QMessageBox.Yes | QMessageBox.No,
                                                                 QMessageBox.Yes) != QMessageBox.Yes:
                return
            f = open(self.wfile.filename(), "wt")
            f.write('COORD W\nPA STANDARD\nCOLOR GREEN\nFONT hershey12\n')
            # source list
            if self.wsel.isChecked():
                sources = [src for src in self.model.sources if src.selected]
            else:
                sources = self.model.sources
            # calculate basis size for crosses (TODO: replace min_size with something more sensible, as this value is in degrees)
            brightnesses = [abs(src.brightness()) for src in sources if src.brightness() != 0]
            min_bright = brightnesses and min(brightnesses)
            min_size = 0.01
            # loop over sources
            busy = BusyIndicator()
            for src in sources:
                ra = src.pos.ra / DEG
                dec = src.pos.dec / DEG
                # figure out source size
                if src.brightness() and min_bright:
                    ysize = (math.log10(abs(src.brightness())) - math.log10(min_bright) + 1) * min_size
                else:
                    ysize = min_size
                xsize = ysize / (math.cos(src.pos.dec) or 1)
                # figure out source style
                style, label = self.model.getSourcePlotStyle(src)
                if style:
                    f.write('# %s\n' % src.name)
                    # write symbol for source
                    f.write('COLOR %s\n' % style.symbol_color)
                    if style.symbol == "plus":
                        f.write('CROSS %.12f %.12f %f %f\n' % (ra, dec, xsize, ysize))
                    elif style.symbol == "cross":
                        f.write('CROSS %.12f %.12f %f %f 45\n' % (ra, dec, ysize, ysize))
                    elif style.symbol == "circle":
                        f.write('CIRCLE %.12f %.12f %f\n' % (ra, dec, ysize))
                    elif style.symbol == "dot":
                        f.write('DOT %.12f %.12f\n' % (ra, dec))
                    elif style.symbol == "square":
                        f.write('CBOX %.12f %.12f %f %f\n' % (ra, dec, xsize, ysize))
                    elif style.symbol == "diamond":
                        f.write('CBOX %.12f %.12f %f %f 45\n' % (ra, dec, xsize, ysize))
                    # write label
                    if label:
                        f.write('FONT hershey%d\n' % (style.label_size * 2))
                        f.write('COLOR %s\n' % style.label_color)
                        f.write('TEXT %.12f %.12f %s\n' % (ra, dec, label))
            f.close()
        except IOError as err:
            busy.reset_cursor()
            self.qerrmsg.showMessage("Error writing Karma annotations file %s: %s" % (filename, str(err)))
            return
        busy.reset_cursor()
        self.parent().showMessage("Wrote Karma annotations for %d sources to file %s" % (len(sources), filename))
        return QDialog.accept(self)


def export_karma_annotations(mainwin, model):
    dialog = getattr(mainwin, '_export_karma_dialog', None)
    if not dialog:
        dialog = mainwin._export_karma_dialog = ExportKarmaDialog(mainwin)
    dialog.setModel(model)
    # show dialog
    return dialog.exec_()


from TigGUI.Tools import registerTool

registerTool("Export Karma annotations...", export_karma_annotations)



# Copyright (C) 2002-2022
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import copy
import math

import numpy
import numpy.ma
from PyQt5.Qt import QObject, QWidget, QHBoxLayout, QLabel, \
    QToolButton, Qt, QColor, QImage, QPixmap, QPainter, QGridLayout, QBrush, QTimer
from PyQt5.Qwt import QwtSlider
from PyQt5.QtCore import pyqtSignal
from scipy.ndimage import measurements

import TigGUI.kitties.utils

_verbosity = TigGUI.kitties.utils.verbosity(name="colormap")
dprint = _verbosity.dprint
dprintf = _verbosity.dprintf


class IntensityMap:
    """An IntensityMap maps a float array into a 0...1 range."""

    def __init__(self, dmin=None, dmax=None):
        """Constructor. An optional data range may be supplied."""
        self.range = None
        self.subset = None
        self.subset_minmax = None
        if dmin is not None:
            if dmax is None:
                raise TypeError("both dmin and dmax must be specified, or neither.""")
            self.setDataRange(dmin, dmax)

    def copy(self):
        return copy.copy(self)

    def setDataRange(self, dmin, dmax):
        """Sets the data range."""
        self.range = dmin, dmax

    def setDataSubset(self, subset, minmax=None):
        """Sets the data subset."""
        self.subset = subset
        self.subset_minmax = minmax

    def getDataSubset(self):
        return self.subset, self.subset_minmax

    def getDataRange(self, data):
        """Returns the set data range, or uses data min/max if it is not set"""
        # use data min/max if no explicit ranges are set
        return self.range or measurements.extrema(data)[:2]

    def remap(self, data):
        """Remaps data into 0...1 range"""
        raise RuntimeError("remap() not implemented in " + str(type(self)))


class LinearIntensityMap(IntensityMap):
    """This scales data linearly between preset min and max values."""

    def remap(self, data):
        d0, d1 = self.getDataRange(data)
        dd = d1 - d0
        if dd:
            return ((data - d0) / dd).clip(0, 1)
        else:
            return numpy.zeros(data.shape, float)


class LogIntensityMap(IntensityMap):
    """This scales data linearly between preset min and max values."""

    def __init__(self, log_cycles=6):
        self.log_cycles = log_cycles

    def remap(self, data):
        # d0,d1 is current data range
        d0, d1 = self.getDataRange(data)
        if d0 == d1:
            return numpy.zeros(data.shape, float)
        dmax = d1 - d0
        data = data - d0
        dmin = dmax * (10 ** (-self.log_cycles))
        # clip data to between dmin and dmax, and take log
        data = numpy.ma.log10(data.clip(dmin, dmax))
        # now rescale
        return (data - math.log10(dmin)) / (math.log10(dmax) - math.log10(dmin))


class HistEqIntensityMap(IntensityMap):
    def __init__(self, nbins=256):
        """Creates intensity mapper which uses histogram equalization."""
        IntensityMap.__init__(self)
        self._nbins = nbins
        self._cdf = self._bins = self.subset = None

    def setDataSubset(self, subset, minmax=None):
        IntensityMap.setDataSubset(self, subset, minmax)
        self._bins = None  # to recompute the CDF

    def setDataRange(self, *range):
        IntensityMap.setDataRange(self, *range)
        self._bins = None  # to recompute the CDF

    def _computeCDF(self, data):
        """Recomputes the CDF using the current data subset and range"""
        dmin, dmax = self.getDataRange(self.subset if self.subset is not None else data)
        if dmin == dmax:
            self._cdf = None
        else:
            dprint(1, "computing CDF for range", dmin, dmax)
            # make cumulative histogram, normalize to 0...1
            hist = measurements.histogram(self.subset if self.subset is not None else data, dmin, dmax, self._nbins)
            cdf = numpy.cumsum(hist)
            if not numpy.all(cdf == 0):
                cdf = cdf / float(cdf[-1])
                # append 0 at beginning, as left side of bin
                self._cdf = numpy.zeros(len(cdf) + 1, float)
                self._cdf[1:] = cdf[...]
                # make array of bin edges
                self._bins = dmin + (dmax - dmin) * numpy.arange(self._nbins + 1) / float(self._nbins)

    def remap(self, data):
        values = None
        if self._bins is None:
            self._computeCDF(data)
        if self._cdf is None:
            return numpy.zeros(data.shape, float)
        if self._bins is not None:
            values = numpy.interp(data.ravel(), self._bins, self._cdf).reshape(data.shape)
        if hasattr(data, 'mask') and values is not None:
            values = numpy.ma.masked_array(values, data.mask)
        if values is not None:
            return values
        else:
            return numpy.zeros(data.shape, float)


class Colormap(QObject):
    """
    A Colormap provides operations for turning normalized float arrays into QImages.
    The default implementation is a linear colormap between two colors.
    """

    def __init__(self, name, color0=QColor("black"), color1=QColor("white"), alpha=(1, 1)):
        QObject.__init__(self)
        self.name = name
        # color is either specified as one argument (which should then be a [3,n] or [4,n] array),
        # or as two QColors orstring names.
        if isinstance(color0, (list, tuple)):
            self._rgb = numpy.array(color0)
            if self._rgb.shape[1] != 3 or self._rgb.shape[0] < 2:
                raise TypeError("expected [N,3] (N>=2) array as first argument")
        else:
            if isinstance(color0, str):
                color0 = QColor(color0)
            if isinstance(color1, str):
                color1 = QColor(color1)
            self._rgb = numpy.array([[color0.red(), color0.green(), color0.blue()],
                                     [color1.red(), color1.green(), color1.blue()]]) / 255.
        self._rgb_arg = numpy.arange(self._rgb.shape[0]) / (self._rgb.shape[0] - 1.0)
        # alpha array
        self._alpha = numpy.array(alpha).astype(float)
        self._alpha_arg = numpy.arange(len(alpha)) / (len(alpha) - 1.0)
        # background brush
        self._brush = None

    def makeQImage(self, width, height):
        data = numpy.zeros((width, height), float)
        data[...] = (numpy.arange(width) / (width - 1.))[:, numpy.newaxis]
        # make brush image -- diag background, with colormap on top
        img = QImage(width, height, QImage.Format_RGB32)
        painter = QPainter(img)
        painter.fillRect(0, 0, width, height, QBrush(QColor("white")))
        painter.fillRect(0, 0, width, height, QBrush(Qt.BDiagPattern))
        painter.drawImage(0, 0, self.colorize(data))
        painter.end()
        return img

    def makeQPixmap(self, width, height):
        data = numpy.zeros((width, height), float)
        data[...] = (numpy.arange(width) / (width - 1.))[:, numpy.newaxis]
        # make brush image -- diag background, with colormap on top
        img = QPixmap(width, height)
        painter = QPainter(img)
        painter.fillRect(0, 0, width, height, QBrush(QColor("white")))
        painter.fillRect(0, 0, width, height, QBrush(Qt.BDiagPattern))
        painter.drawImage(0, 0, self.colorize(data))
        painter.end()
        return img

    def makeBrush(self, width, height):
        return QBrush(self.makeQImage(width, height))

    def colorize(self, data, alpha=None):
        """Converts normalized data (0...1) array into a QImage of the same dimensions.
        'alpha', if set, is a 0...1 array of the same size, which is mapped to the alpha channel
        (i.e. 0 for fully transparent and 1 for fully opaque).
        If data is a masked array, masked pixels will be fully transparent."""
        # setup alpha channel
        if alpha is None:
            alpha = numpy.interp(data.ravel(), self._alpha_arg, self._alpha).reshape(data.shape)
        alpha = numpy.round(255 * alpha).astype(numpy.int32).clip(0, 255)
        # make RGB arrays
        rgbs = [(numpy.interp(data.ravel(), self._rgb_arg, self._rgb[:, i]).
                 reshape(data.shape) * 255).round().astype(numpy.int32).clip(0, 255)
                for i in range(3)]
        # add data mask
        mask = getattr(data, 'mask', None)
        if mask is not None and mask is not False:
            alpha[mask] = 0
            for x in rgbs:
                x[mask] = 0
        # do the deed
        return self.QARGBImage(alpha, *rgbs)

    def makeControlWidgets(self, parent):
        """Creates control widgets for the colormap's internal parameters.
        "parent" is a parent widget.
        Returns None if no controls are required"""
        return None

    class QARGBImage(QImage):
        """This is a QImage which is constructed from an A,R,G,B arrays."""

        def __init__(self, a, r, g, b):
            nx, ny = r.shape
            argb = (a << 24) | (r << 16) | (g << 8) | b
            # transpose array, as it is in column-major (C order), while QImages are in row-major order
            dprint(5, "making qimage of size", nx, ny)
            self._buffer = argb.transpose().tostring()
            QImage.__init__(self, self._buffer, nx, ny, QImage.Format_ARGB32)


class ColormapWithControls(Colormap):
    """This is a base class for a colormap with controls knobs"""
    colormapChanged = pyqtSignal()
    colormapPreviewed = pyqtSignal()

    class SliderControl(QObject):
        """This class implements a slider control for a colormap"""
        valueChanged = pyqtSignal(float)
        valueMoved = pyqtSignal(float)

        def __init__(self, name, value, minval, maxval, step, format="%s: %.1f"):
            QObject.__init__(self)
            self.name, self.value, self.minval, self.maxval, self.step, self.format = \
                name, value, minval, maxval, step, format
            self._default = value
            self._wlabel = None
            self._wreset = None
            self._wslider = None
            self._wslider_timer = None

        def makeControlWidgets(self, parent, gridlayout, row, column):
            toprow = QWidget(parent)
            gridlayout.addWidget(toprow, row * 2, column)
            top_lo = QHBoxLayout(toprow)
            top_lo.setContentsMargins(0, 0, 0, 0)
            self._wlabel = QLabel(self.format % (self.name, self.value), toprow)
            top_lo.addWidget(self._wlabel)
            self._wreset = QToolButton(toprow)
            self._wreset.setText("reset")
            self._wreset.setToolButtonStyle(Qt.ToolButtonTextOnly)
            self._wreset.setAutoRaise(True)
            self._wreset.setEnabled(self.value != self._default)
            self._wreset.clicked.connect(self._resetValue)
            top_lo.addWidget(self._wreset)
            self._wslider = QwtSlider(parent)
            self._wslider.setOrientation(Qt.Horizontal)
            # This works around a stupid bug in QwtSliders -- see comments on histogram zoom wheel above
            self._wslider_timer = QTimer(parent)
            self._wslider_timer.setSingleShot(True)
            self._wslider_timer.setInterval(500)
            self._wslider_timer.timeout.connect(self.setValue)
            gridlayout.addWidget(self._wslider, row * 2 + 1, column)
            self._wslider.setScale(self.minval, self.maxval)
            # self._wslider.setScaleStepSize(self.step)
            self._wslider.setValue(self.value)
            self._wslider.setTracking(False)
            self._wslider.valueChanged.connect(self.setValue)
            self._wslider.sliderMoved.connect(self._previewValue)

        def _resetValue(self):
            self._wslider.setValue(self._default)
            self.setValue(self._default)

        def setValue(self, value=None, notify=True):
            # only update widgets if already created
            self.value = value
            if self._wlabel is not None:
                if value is None:
                    self.value = value = self._wslider.value()
                self._wreset.setEnabled(value != self._default)
                self._wlabel.setText(self.format % (self.name, self.value))
                # stop timer if being called to finalize the change in value
                if notify:
                    self._wslider_timer.stop()
                    self.valueChanged.emit(self.value)

        def _previewValue(self, value):
            self.setValue(notify=False)
            self._wslider_timer.start(500)
            self.valueMoved.emit(self.value)

    def emitChange(self, *dum):
        self.colormapChanged.emit()

    def emitPreview(self, *dum):
        self.colormapPreviewed.emit()

    def loadConfig(self, config):
        pass

    def saveConfig(self, config, save=True):
        pass


class CubeHelixColormap(ColormapWithControls):
    """This implements the "cubehelix" colour scheme proposed by Dave Green:
    D. Green 2011, Bull. Astr. Soc. India (2011) 39, 289–295
    http://arxiv.org/pdf/1108.5083v1
    """

    colormapChanged = pyqtSignal()

    def __init__(self, gamma=1, rgb=0.5, rots=-1.5, hue=1.2, name="CubeHelix"):
        ColormapWithControls.__init__(self, name)
        self.gamma = self.SliderControl("Gamma", gamma, 0, 6, .1)
        self.color = self.SliderControl("Colour", rgb, 0, 3, .1)
        self.cycles = self.SliderControl("Cycles", rots, -10, 10, .1)
        self.hue = self.SliderControl("Hue", hue, 0, 2, .1)

    def colorize(self, data, alpha=None):
        """Converts normalized data (0...1) array into a QImage of the same dimensions.
        'alpha', if set, is a 0...1 array of the same size, which is mapped to the alpha channel
        (i.e. 0 for fully transparent and 1 for fully opaque).
        If data is a masked array, masked pixels will be fully transparent."""
        # setup alpha channel
        if alpha is None:
            alpha = numpy.zeros(data.shape, dtype=numpy.int32)
            alpha[...] = 255
        else:
            alpha = numpy.round(255 * alpha).astype(numpy.int32).clip(0, 255)
        # make RGB arrays
        dg = data ** self.gamma.value
        a = self.hue.value * dg * (1 - dg) / 2
        phi = 2 * math.pi * (self.color.value / 3 + self.cycles.value * data)
        cosphi = a * numpy.cos(phi)
        sinphi = a * numpy.sin(phi)
        r = dg - 0.14861 * cosphi + 1.78277 * sinphi
        g = dg - 0.29227 * cosphi - 0.90649 * sinphi
        b = dg + 1.97249 * cosphi
        rgbs = [(x * 255).round().astype(numpy.int32).clip(0, 255) for x in (r, g, b)]
        # add data mask
        mask = getattr(data, 'mask', None)
        if mask is not None and mask is not False:
            alpha[mask] = 0
            for x in rgbs:
                x[mask] = 0
        # do the deed
        return self.QARGBImage(alpha, *rgbs)

    def makeControlWidgets(self, parent):
        """Creates control widgets for the colormap's internal parameters.
        "parent" is a parent widget.
        Returns None if no controls are required"""
        top = QWidget(parent)
        layout = QGridLayout(top)
        layout.setContentsMargins(0, 0, 0, 0)
        for irow, icol, control in ((0, 0, self.gamma), (0, 1, self.color), (1, 0, self.cycles), (1, 1, self.hue)):
            control.makeControlWidgets(top, layout, irow, icol)
            control.valueChanged.connect(self.emitChange)
            control.valueMoved.connect(self.emitPreview)
        return top

    def loadConfig(self, config):
        for name in "gamma", "color", "cycles", "hue":
            control = getattr(self, name)
            value = config.getfloat("cubehelix-colourmap-%s" % name, control.value)
            control.setValue(value, notify=False)

    def saveConfig(self, config, save=True):
        for name in "gamma", "color", "cycles", "hue":
            control = getattr(self, name)
            config.set("cubehelix-colourmap-%s" % name, control.value, save=save)


# instantiate "static" colormaps (i.e. those that have no internal parameters, and thus can be
# shared among images without instantiating a new Colormap object for each)
GreyscaleColormap = Colormap("Greyscale")
TransparentFuchsiaColormap = Colormap("Transparent Fuchsia", color0="fuchsia", color1="fuchsia", alpha=(0, 1))

from .ColormapTables import Karma

_karma_colormaps = [
    Colormap(cmap, getattr(Karma, cmap))
    for cmap in [
        "Background",
        "Heat",
        "Isophot",
        "Mousse",
        "Rainbow",
        "RGB",
        "RGB2",
        "Smooth",
        "Staircase",
        "Mirp",
        "Random"]
]


def getColormapList():
    """Returns list of Colormap instances."""

    # Some colormaps need a unique instantiation (because they have parameters)
    # For the rest, use the static objects
    return [GreyscaleColormap,
            CubeHelixColormap(),
            TransparentFuchsiaColormap] + _karma_colormaps
